import math
from typing import List, Tuple


def angle_between_vectors(v1: Tuple[float, float], v2: Tuple[float, float]) -> float:
    """

    Returns the angle in radians between two vectors 'v1' and 'v2'

    Parameters
    ----------
    v1 : (float, float)
        First vector as tuple
    v2 : (float, float)
        Second vector as tuple

    Returns
    -------
    angle : float
        Angle between v1 and v2 in radians

    """
    dot = v1[0] * v2[0] + v1[1] * v2[1]
    det = v1[0] * v2[1] - v1[1] * v2[0]
    angle = math.atan2(det, dot)
    return angle


def distance_v1_v2(x1: float, y1: float, x2: float, y2: float) -> float:
    """
    Returns the distance of a two-dimensional point.

    Parameters
    ----------
    x1 : float
        First coordinate of the first point
    y1 : float
        Second coordinate of the first point
    x2 : float
        First coordinate of the second point
    y2 : float
        Second coordinate of the second point

    Returns
    -------
    distance : float
        Distance between two-dimensional coordinate (x1, y1) and (x2, y2)
    """
    dist_x = x1 - x2
    dist_y = y1 - y2
    dist = math.sqrt(dist_x**2 + dist_y**2)
    return dist


def quadratic_bezier(
    dt: float, p0: Tuple[float, float], p1: Tuple[float, float], p2: Tuple[float, float]
) -> Tuple[float, float]:
    """
    Calculates a quadractic Bezier-curve.

    Args:
        t: Current time step
        p0: Point 0 of the Bezier curve
        p1: Point 1 of the Bezier curve
        p2: Point 2 of the Bezier curve

    Returns:
        The quadractic Bezier curve
    """
    # Calculate the coordinates of the point on the curve at parameter value t
    x = (1 - dt) ** 2 * p0[0] + 2 * (1 - dt) * dt * p1[0] + dt**2 * p2[0]
    y = (1 - dt) ** 2 * p0[1] + 2 * (1 - dt) * dt * p1[1] + dt**2 * p2[1]
    return (x, y)


def rotate_points_around_origin(points: List, theta: float) -> List[Tuple[float, float]]:
    """
    Rotates a 1D-list around its first point by the given angle.

    Args:
        points: List of values to rotate
        theta: Rotation angle

    Returns:
        Rotated list of values
    """
    theta = math.radians(theta)  # convert theta to radians
    theta_cos, theta_sin = math.cos(theta), math.sin(theta)
    rotated_points = []
    for point in points:
        x, y = point
        rotated_points.append((x * theta_cos - y * theta_sin, x * theta_sin + y * theta_cos))
    return rotated_points


def rotate_points_around_index(points: List, theta: float, index: int) -> List[Tuple[float, float]]:
    """
    Rotates a 1D-list around a given index by the given angle.


    Args:
        points: List of valeus to rotate
        theta: Rotation angle

    Returns:
        Rotated list of values
    """
    # get point at index
    x_at_index = points[index][0]
    y_at_index = points[index][1]

    # translate points to the origin
    points = [(x - x_at_index, y - y_at_index) for x, y in points]

    # rotate the points around the origin
    rotated_points = rotate_points_around_origin(points, theta)

    # translate it back
    rotated_points = [(x + x_at_index, y + y_at_index) for x, y in rotated_points]

    return rotated_points


def rotate_points_around_mid_point(points: list, theta: float) -> List[Tuple[float, float]]:
    """
    Rotates a 1D-list around its mid-point by a given angle.

    Args:
        points: List of values to rotate
        theta: Rotation angle

    Returns:
        Rotated list of values
    """
    # find the midpoint of the points
    x_mid = sum(x for x, _ in points) / len(points)
    y_mid = sum(y for _, y in points) / len(points)

    # translate the points to the origin
    points = [(x - x_mid, y - y_mid) for x, y in points]

    # rotate the points around the origin
    rotated_points = rotate_points_around_origin(points, theta)

    # translate the points back to the midpoint
    rotated_points = [(x + x_mid, y + y_mid) for x, y in rotated_points]

    return rotated_points



from dataclasses import dataclass
from typing import Any, Dict, Tuple

import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import OptimizeResult

from flying_discs.frispy.scipy_backend.eom import EOM
from flying_discs.frispy.scipy_backend.model import Model


@dataclass
class Disc:
    # pylint: disable=too-many-instance-attributes
    """Flying spinning disc object.

    One can specify initial coordinates, initial kinematic variables, the two
    non-zero elements of the moments of inertia tensor, the face area, mass,
    air density, and gravitational acceleration. Note that the default area,
    mass, and inertial moments are for Discraft Ultrastars (0.175 kg).

    MKS units used throughout. Angular units use radians (rad), and
    angular velocities are in rad/s.

    Args:
        x: horizontal position
        y: horizontal position
        z: vertical position
        vx: horizontal speed
        vy: horizontal speed
        vz: vertical speed
        phi: first Euler angle. Pitch angle when velocity is horizontal
        theta: second Euler angle. Roll angle when velocity is horizontal
        gamma: third Euler angle. Spin angle
        dphi: angular velocity
        dtheta: angular velocity
        dgamma: spin angular velocity
        area: face area
        I_xx: pitch and roll moments of inertia
        I_zz: spin moment of inertia
        air_density: default is sea level
        g: gravitational acceleration
        model: defines how accelerations depend on velocities
        eom: equations of motion
    """

    def __init__(
        self,
        area: float = 0.058556,
        I_xx: float = 0.001219,
        I_zz: float = 0.002352,
        mass: float = 0.175,
        air_density: float = 1.225,
        g: float = 9.81,
    ) -> None:
        self.area = area
        self.I_xx = I_xx
        self.I_zz = I_zz
        self.mass = mass
        self.air_density = air_density
        self.g = g
        self.model = Model()
        self.eom = EOM(
            area=self.area,
            I_xx=self.I_xx,
            I_zz=self.I_zz,
            mass=self.mass,
            air_density=self.air_density,
            g=self.g,
        )

    def compute_trajectory(
        self,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 1.0,
        vx: float = 10.0,
        vy: float = 0.0,
        vz: float = 0.0,
        phi: float = 0.0,
        theta: float = 0.0,
        gamma: float = 0.0,
        dphi: float = 0.0,
        dtheta: float = 0.0,
        dgamma: float = 62.0,
        flight_time: float = 3.0,
        n_times: int = 100,
        **solver_kwargs: Any,
    ) -> Tuple[Dict[str, np.ndarray], OptimizeResult]:
        # pylint: disable=too-many-positional-arguments, too-many-arguments, too-many-locals
        """Call the differential equation solver to compute the trajectory.

        The kinematic variables and timesteps are saved
        as the `current_trajectory` attribute, which is a dictionary,
        which is also returned by this function.

        See `the scipy docs
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html#scipy.integrate.solve_ivp>`_
        for more information on the solver.

        .. warning::

           `flight_time` is ignored if `t_span` is a key in
           `solver_kwargs`.

        Args:
            flight_time: time in seconds that the simulation
                will run over
            n_times: Number of samples in time for the trajectory.
                Samples are spaced evenly in time from ``(0, flight_time)``.
            solver_kwargs: extra keyword arguments to pass
                to the :meth:`scipy.integrate.solve_ivp`

        Returns:
            first element is a dict of all kinematic variables and the
            second is the ``solve_ivp`` results object
        """
        # Pop out these kwargs and take defaults based on our API
        t_span: tuple[float, float] = solver_kwargs.pop("t_span", (0, flight_time))
        t_eval = solver_kwargs.pop("t_eval", np.linspace(t_span[0], t_span[1], n_times))

        # Call the solver
        result = solve_ivp(
            fun=self.eom.compute_derivatives,
            t_span=t_span,
            y0=[
                x,
                y,
                z,
                vx,
                vy,
                vz,
                phi,
                theta,
                gamma,
                dphi,
                dtheta,
                dgamma,
            ],
            t_eval=t_eval,
            **solver_kwargs,
        )

        return (
            {
                "times": result.t,
                "x": result.y[0],
                "y": result.y[1],
                "z": result.y[2],
                "vx": result.y[3],
                "vy": result.y[4],
                "vz": result.y[5],
                "phi": result.y[6],
                "theta": result.y[7],
                "gamma": result.y[8],
                "dphi": result.y[9],
                "dtheta": result.y[10],
                "dgamma": result.y[11],
            },
            result,
        )



from typing import Any

import equinox as eqx
import jax
import jax.numpy as jnp

from flying_discs.frispy.jax_backend.model import Model


class _GeometricQuantitiesResult(eqx.Module):
    xhat: jnp.ndarray
    yhat: jnp.ndarray
    zhat: jnp.ndarray
    angle_of_attack: jnp.ndarray
    rotation_matrix: jnp.ndarray
    w_prime: jnp.ndarray
    w_lab: jnp.ndarray
    w: jnp.ndarray


class _ForcesResult(eqx.Module):
    lift: jnp.ndarray
    drag: jnp.ndarray
    grav: jnp.ndarray
    mass: float

    @property
    def total(self) -> jnp.ndarray:
        return self.lift + self.drag + self.grav

    @property
    def acc(self) -> jnp.ndarray:
        return self.total / self.mass


class _TorquesResult(eqx.Module):
    x_lab: jnp.ndarray
    y_lab: jnp.ndarray
    z: jnp.ndarray
    x: jnp.ndarray
    y: jnp.ndarray

    @property
    def total(self) -> jnp.ndarray:
        return self.x + self.y + self.z


# -------------------------------------------------------------------
# Equations of Motion (EOM)
# -------------------------------------------------------------------


class EOM(eqx.Module):
    # pylint: disable=too-many-instance-attributes
    """Equations of motion for a disc (JAX compatible)."""

    area: float
    diameter: jnp.ndarray
    I_xx: float
    I_zz: float
    mass: float
    air_density: float
    g: float
    model: Model
    force_per_v2: float
    torque_per_v2: jnp.ndarray
    grav: jnp.ndarray
    z_hat: jnp.ndarray

    def __init__(
        self,
        area: float,
        I_xx: float,
        I_zz: float,
        mass: float,
        air_density: float = 1.225,
        g: float = 9.81,
        model: Model = Model(),
    ):
        self.area = area
        self.diameter = 2 * jnp.sqrt(self.area / jnp.pi)
        self.I_xx = I_xx
        self.I_zz = I_zz
        self.mass = mass
        self.model = model
        self.air_density = air_density
        self.g = g
        self.force_per_v2 = 0.5 * self.air_density * self.area
        self.torque_per_v2 = self.force_per_v2 * self.diameter
        self.grav = jnp.array([0.0, 0.0, -self.mass * self.g])
        self.z_hat = jnp.array([0.0, 0.0, 1.0])

    @classmethod
    def rotation_matrix_from_phi_theta(cls, phi: float, theta: float) -> jnp.ndarray:
        sp, cp = jnp.sin(phi), jnp.cos(phi)
        st, ct = jnp.sin(theta), jnp.cos(theta)
        return cls.rotation_matrix(sp, cp, st, ct)

    @staticmethod
    def rotation_matrix(sp: jnp.ndarray, cp: jnp.ndarray, st: jnp.ndarray, ct: jnp.ndarray) -> jnp.ndarray:
        return jnp.array([[ct, sp * st, -st * cp], [0.0, cp, sp], [st, -sp * ct, cp * ct]])

    @classmethod
    def compute_angle_of_attack(
        cls,
        phi: jnp.ndarray,
        theta: jnp.ndarray,
        velocity: jnp.ndarray,
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        sp, cp = jnp.sin(phi), jnp.cos(phi)
        st, ct = jnp.sin(theta), jnp.cos(theta)
        rotation_matrix = cls.rotation_matrix(sp, cp, st, ct)
        zhat = rotation_matrix[2]
        v_dot_zhat = velocity @ zhat
        v_in_plane = velocity - zhat * v_dot_zhat
        angle_of_attack = -jnp.arctan(v_dot_zhat / jnp.linalg.norm(v_in_plane))
        # pylint: disable=duplicate-code
        return (
            angle_of_attack,
            sp,
            cp,
            st,
            ct,
            rotation_matrix,
            v_dot_zhat,
            v_in_plane,
        )

    def geometric_quantities(
        self,
        phi: jnp.ndarray,
        theta: jnp.ndarray,
        velocity: jnp.ndarray,
        angular_velocity: jnp.ndarray,
    ) -> _GeometricQuantitiesResult:
        # pylint: disable=too-many-locals
        (
            angle_of_attack,
            _,
            _,
            st,
            ct,
            rotation_matrix,
            _,
            v_in_plane,
        ) = self.compute_angle_of_attack(phi, theta, velocity)

        zhat: jnp.ndarray = rotation_matrix[2]
        xhat: jnp.ndarray = v_in_plane / jnp.linalg.norm(v_in_plane)
        yhat = jnp.cross(zhat, xhat)
        w_prime = jnp.array(
            [
                angular_velocity[0] * ct,
                angular_velocity[1],
                angular_velocity[0] * st + angular_velocity[2],
            ]
        )
        w_lab = w_prime @ rotation_matrix
        w = jnp.array([xhat, yhat, zhat]) @ w_lab
        # pylint: disable=duplicate-code
        return _GeometricQuantitiesResult(
            xhat=xhat,
            yhat=yhat,
            zhat=zhat,
            angle_of_attack=angle_of_attack,
            rotation_matrix=rotation_matrix,
            w_prime=w_prime,
            w_lab=w_lab,
            w=w,
        )

    def compute_forces(
        self,
        velocity: jnp.ndarray,
        geometric_quantities: _GeometricQuantitiesResult,
    ) -> _ForcesResult:
        vhat = velocity / jnp.linalg.norm(velocity)
        force_amplitude = self.force_per_v2 * (velocity @ velocity)
        res = _ForcesResult(
            lift=self.model.lift(geometric_quantities.angle_of_attack)
            * force_amplitude
            * jnp.cross(vhat, geometric_quantities.yhat),
            drag=self.model.drag(geometric_quantities.angle_of_attack) * force_amplitude * (-vhat),
            grav=self.grav,
            mass=self.mass,
        )
        return res

    def compute_torques(
        self,
        velocity: jnp.ndarray,
        geometric_quantities: _GeometricQuantitiesResult,
    ) -> _TorquesResult:
        torque_amplitude = self.torque_per_v2 * (velocity @ velocity)
        wx, wy, wz = geometric_quantities.w
        t_x_lab = self.model.x(wx, wz) * torque_amplitude * geometric_quantities.xhat
        t_y_lab = self.model.y(geometric_quantities.angle_of_attack, wy) * torque_amplitude * geometric_quantities.yhat
        z = self.model.z(wz) * torque_amplitude * self.z_hat
        x = geometric_quantities.rotation_matrix @ t_x_lab
        y = geometric_quantities.rotation_matrix @ t_y_lab
        return _TorquesResult(x_lab=t_x_lab, y_lab=t_y_lab, z=z, x=x, y=y)

    def compute_derivatives(self, _: jnp.number, coordinates: jnp.ndarray, __: Any) -> jnp.ndarray:
        """Right hand side of the ordinary differential equations (Diffrax compatible)."""

        def calculate_physics(coords: jnp.ndarray) -> jnp.ndarray:
            velocity = coords[3:6]
            ang_velocity = coords[9:12]
            geometric_quantities = self.geometric_quantities(coords[6], coords[7], velocity, ang_velocity)
            forces = self.compute_forces(velocity, geometric_quantities)
            torques = self.compute_torques(velocity, geometric_quantities)
            return jnp.concatenate((velocity, forces.acc, ang_velocity, torques.total))

        def frozen_state(coords: jnp.ndarray) -> jnp.ndarray:
            return jnp.zeros_like(coords)

        # Replaces `if coordinates[2] <= 0:` for JAX tracing compatibility
        y0: jnp.ndarray = jax.lax.cond(coordinates[2] <= 0, frozen_state, calculate_physics, coordinates)
        return y0



from abc import ABC


class Constants(ABC):
    # pylint: disable = invalid-name
    # defaults to ultrastar and see-level
    AREA: float = 0.058556  # m ^ 2
    I_xx: float = 0.001219  # kg * m ^ 2
    I_zz: float = 0.002352  # kg * m ^ 2
    MASS: float = 0.175  # kg
    RH0: float = 1.225  # air density, kg / m ^ 3
    GRAVITY: float = 9.81  # m / s ^ 2


class FrispyUltrastarConstants(Constants):
    AREA: float = 0.058556  # m ^ 2
    I_xx: float = 0.001219  # kg * m ^ 2
    I_zz: float = 0.002352  # kg * m ^ 2
    MASS: float = 0.175  # kg



from dataclasses import dataclass

import numpy as np


@dataclass
class Model:
    # pylint: disable=too-many-instance-attributes
    """Kinematic model for a disc.

    Holds all of the aerodynamic parameters coupling the kinematic variables
    (velocities, and angular velocities) to the force and torque magnitudes.

    Args:
        PL0: lift force at zero angle of attack (aoa; i.e. the intercept)
        PLa: lift force linear scaling of aoa
        PD0: drag force at zero aoa (i.e. the intercept)
        PDa: drag force linear scaling of aoa
        PTxwx: torque around x-hat linear scaling with x-hat angular velocity
        PTxwz: torque around x-hat linear scaling with z-hat angular velocity (spin)
        PTy0: torque around y-hat intercept
        PTya: torque around y-hat linear scaling with aoa
        PTywy: torque around y-hat linear scaling with y-hat angular velocity
        PTzwz: torque around z-hat linear scaling with z-hat angular velocity (spin)
        alpha_0: angle of attack at minimal drag (i.e. critical aoa)
    """

    PL0: float = 0.33
    PLa: float = 1.9
    PD0: float = 0.18
    PDa: float = 0.69
    PTxwx: float = -0.013
    PTxwz: float = -0.0017
    PTy0: float = -0.082
    PTya: float = 0.43
    PTywy: float = -0.014
    PTzwz: float = -0.000034
    alpha_0: float = 4 * np.pi / 180

    #####################################################################
    # Below are functions connecting physical variables to force/torque #
    # scaling factors (the `C`s)                                        #
    #####################################################################

    def lift(self, alpha: float) -> float:
        """Lift force scale factor.

        Linear in the angle of attack (``alpha``).

        Args:
            alpha: angle of attack in radians

        Returns:
            lift force scale factor
        """
        return self.PL0 + self.PLa * alpha

    def drag(self, alpha: float) -> float:
        """Drag force scale factor.

        Quadratic in the angle of attack (``alpha``).

        Args:
            alpha: angle of attack in radians

        Returns:
             drag force scale factor
        """
        return self.PD0 + self.PDa * (alpha - self.alpha_0) ** 2

    def x(self, wx: float, wz: float) -> float:
        """Torque scale factor around x-hat.

        Linearly in the z angular velocity
        (``w_z``) and the x angular velocity (``w_x``).

        Args:
            wx: x angular velocity in radians per second
            wz: z angular velocity in radians per second

        Returns:
            torque scale factor around x-hat
        """
        return self.PTxwx * wx + self.PTxwz * wz

    def y(self, alpha: float, wy: float) -> float:
        """Torque scale factor around y-hat.

        Linearly in the y angular velocity
        (``w_y``) and the angle of attack (``alpha``).

        Args:
            alpha: angle of attack in radians
            wy: y angular velocity in radians per second

        Returns:
            torque scale factor around y-hat
        """
        return self.PTy0 + self.PTywy * wy + self.PTya * alpha

    def z(self, wz: float) -> float:
        """Torque scale factor around z-hat.

        Linear in the z angular velocity (``w_z``).

        Args:
            wz: z angular velocity in radians per second

        Returns:
            torque scale factor around z-hat
        """
        return self.PTzwz * wz



import math
from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np

from flying_discs.morrison.base import MorrisonBaseCalculator
from flying_discs.morrison.constants import Constants
from flying_discs.morrison.coordinates import (
    MorrisonPosition2D,
    MorrisonPosition3D,
    MorrisonTrajectory2D,
    MorrisonTrajectory3D,
)
from flying_discs.utils import angle_between_vectors, distance_v1_v2


@dataclass
class MorrisonLinearThrow:
    # pylint: disable=too-many-instance-attributes
    trajectory: MorrisonTrajectory3D
    base_trajectory: MorrisonTrajectory2D = field(compare=False)
    constants: Constants
    initial_position: MorrisonPosition3D
    v0: float
    angle_of_attack: float
    direction_angle: float
    deltaT: float
    target_x: Optional[float] = None
    target_y: Optional[float] = None


class MorrisonLinearCalculator:
    def __init__(self, constants: Constants) -> None:
        self._base_calculator = MorrisonBaseCalculator(constants)
        self.constants = constants

    def calculate_trajectory(
        self,
        initial_position: MorrisonPosition3D,
        v0: float,
        angle_of_attack: float,
        direction_angle: float,
        deltaT: float,
    ) -> MorrisonLinearThrow:
        base_trajectory = self._base_calculator.calculate_trajectory(
            initial_position.z, v0, angle_of_attack, deltaT
        ).trajectory
        linear_trajectory: List[MorrisonPosition3D] = []
        for i, base_position in enumerate(base_trajectory):
            if i == 0:
                linear_trajectory.append(
                    self._calculate_initial_trajectory_step(initial_position, base_position, direction_angle)
                )
                continue
            linear_trajectory.append(
                self._calculate_next_trajectory_step(base_position, linear_trajectory[i - 1], direction_angle, deltaT)
            )
        return MorrisonLinearThrow(
            MorrisonTrajectory3D(linear_trajectory),
            base_trajectory,
            self.constants,
            initial_position,
            v0,
            angle_of_attack,
            direction_angle,
            deltaT,
        )

    @staticmethod
    def _calculate_initial_trajectory_step(
        initial_position: MorrisonPosition3D, base_position: MorrisonPosition2D, direction_angle: float
    ) -> MorrisonPosition3D:
        return MorrisonPosition3D(
            x=initial_position.x,
            y=initial_position.y,
            z=initial_position.z,
            vx=base_position.vx * math.cos(direction_angle),
            vy=base_position.vx * math.sin(direction_angle),
            vz=base_position.vz,
            ax=base_position.ax * math.cos(direction_angle),
            ay=base_position.ax * math.sin(direction_angle),
            az=base_position.az,
        )

    @staticmethod
    def _calculate_next_trajectory_step(
        base_position: MorrisonPosition2D, previous_position: MorrisonPosition3D, direction_angle: float, deltaT: float
    ) -> MorrisonPosition3D:
        new_ax = base_position.ax * math.cos(direction_angle)
        new_ay = base_position.ax * math.sin(direction_angle)
        new_vx = previous_position.vx + new_ax
        new_vy = previous_position.vy + new_ay
        new_x = previous_position.x + new_vx * deltaT
        new_y = previous_position.y + new_vy * deltaT
        return MorrisonPosition3D(
            x=new_x,
            y=new_y,
            z=base_position.z,
            vx=new_vx,
            vy=new_vy,
            vz=base_position.vz,
            ax=new_ax,
            ay=new_ay,
            az=base_position.az,
        )

    def _find_trajectory_to_position(
        self,
        initial_position: MorrisonPosition3D,
        angle_of_attack: float,
        direction_angle: float,
        dist: float,
        deltaT: float,
    ) -> tuple[float, MorrisonTrajectory3D, MorrisonTrajectory2D]:
        v0 = 0.0
        distance_traveled = -math.inf
        height_at_target = -math.inf
        while distance_traveled < dist or height_at_target < 0:
            # TODO: Add exit condition if no solution is found after a certain number of iterations
            v0 += 0.1
            distance_traveled = -math.inf
            base_throw = self.calculate_trajectory(initial_position, v0, angle_of_attack, direction_angle, deltaT)
            trajectory = base_throw.trajectory
            distance_traveled = float(
                np.linalg.norm(
                    [trajectory[-1].x, trajectory[-1].y],
                )
            )
            # TODO: Handle non zero for initial x/y position on non-zero
            # See: https://github.com/crzdg/flying-discs/pull/16#discussion_r2991607092
            d_ = np.array([np.linalg.norm([p.x, p.y]) for p in trajectory])
            idx = np.where(d_ > dist)[0]
            if idx.size > 0:
                start_idx = idx[0]
                height_at_target = trajectory[start_idx].z
        return v0, trajectory, base_throw.base_trajectory

    def calculate_trajectory_to_position(
        self,
        initial_position: MorrisonPosition3D,
        angle_of_attack: float,
        target_x: float,
        target_y: float,
        deltaT: float,
    ) -> MorrisonLinearThrow:
        direction_angle = angle_between_vectors((1, 0), (target_x - initial_position.x, target_y - initial_position.y))
        dist = distance_v1_v2(target_x, target_y, initial_position.x, initial_position.y)
        found_v0, trajectory, base_trajectory = self._find_trajectory_to_position(
            initial_position, angle_of_attack, direction_angle, dist, deltaT
        )
        return MorrisonLinearThrow(
            trajectory,
            base_trajectory,
            self.constants,
            initial_position,
            found_v0,
            angle_of_attack,
            direction_angle,
            deltaT,
            target_x,
            target_y,
        )



from dataclasses import dataclass

import numpy as np

from flying_discs.frispy.scipy_backend.model import Model


@dataclass
class _GeometricQuantitiesResult:
    xhat: np.ndarray
    yhat: np.ndarray
    zhat: np.ndarray
    angle_of_attack: float
    rotation_matrix: np.ndarray
    w_prime: np.ndarray
    w_lab: np.ndarray
    w: np.ndarray


@dataclass
class _ForcesResult:
    lift: np.ndarray
    drag: np.ndarray
    grav: np.ndarray
    mass: float

    @property
    def total(self) -> np.ndarray:
        total: np.ndarray = self.lift + self.drag + self.grav
        return total

    @property
    def acc(self) -> np.ndarray:
        acc: np.ndarray = self.total / self.mass
        return acc


@dataclass
class _TorquesResult:
    x_lab: np.ndarray
    y_lab: np.ndarray
    z: np.ndarray
    x: np.ndarray
    y: np.ndarray

    @property
    def total(self) -> np.ndarray:
        total: np.ndarray = self.x + self.y + self.z
        return total


class EOM:
    # pylint: disable=too-many-instance-attributes
    """Equations of motion for a disc.

    ``EOM`` is short for "equations of motion". Used to run the ODE solver
    from ``scipy``. It takes in a model for the disc, the trajectory object,
    the environment, and implements the functions for calculating forces
    and torques.

    Args:
        area: disc area
        I_xx: pitch and roll moments of inertia
        I_zz: spin moment of inertia
        mass: of the disc
        air_density: thiccness of the air
        g: gravitational acceleration
        model: must yield force and torque coefficients
    """

    def __init__(
        self,
        area: float,
        I_xx: float,
        I_zz: float,
        mass: float,
        air_density: float = 1.225,
        g: float = 9.81,
        # TODO: State handling of the model.
        # Also see constants.py
        # SEE: https://github.com/crzdg/flying-discs/pull/16#discussion_r2991607292
        model: Model = Model(),
    ):
        """Constructor."""
        self.area = area
        self.diameter = 2 * np.sqrt(self.area / np.pi)
        self.I_xx = I_xx
        self.I_zz = I_zz
        self.mass = mass
        self.model = model
        self.air_density = air_density
        self.g = g
        # Pre-compute some values to optimize the ODEs
        self.force_per_v2 = 0.5 * self.air_density * self.area  # N / (m/s)^2
        self.torque_per_v2 = self.force_per_v2 * self.diameter  # N * m / (m/s)^2
        self.grav = np.array([0, 0, -self.mass * self.g])
        self.z_hat = np.array([0, 0, 1])

    @classmethod
    def rotation_matrix_from_phi_theta(cls, phi: float, theta: float) -> np.ndarray:
        """Rotation matrix."""
        sp, cp = np.sin(phi), np.cos(phi)
        st, ct = np.sin(theta), np.cos(theta)
        return cls.rotation_matrix(sp, cp, st, ct)

    @staticmethod
    def rotation_matrix(sp: float, cp: float, st: float, ct: float) -> np.ndarray:
        """Compute the rotation matrix.

        Compute the (partial) rotation matrix that transforms from the
        lab frame to the disc frame. Note that because of azimuthal
        symmetry, the azimuthal angle (`gamma`) is not used.

        This matrix (R) can be used to transform a vector from the lab frame (L)
        into the disk frame (D), i.e.: r_D = R dot r_L.

        The ``z_hat`` unit vector in the disk frame (D) will always be pointing
        perpendicular up from the top face of the disk.
        """
        return np.array([[ct, sp * st, -st * cp], [0, cp, sp], [st, -sp * ct, cp * ct]])

    @classmethod
    def compute_angle_of_attack(
        cls,
        phi: float,
        theta: float,
        velocity: np.ndarray,
    ) -> tuple[float, float, float, float, float, np.ndarray, float, np.ndarray]:
        """Compute the angle of attack."""
        # TODO: Does not guard against zhat (norm == 0)
        # SEE: https://github.com/crzdg/flying-discs/pull/16#discussion_r2991607301
        # Rotation matrix
        sp, cp = np.sin(phi), np.cos(phi)
        st, ct = np.sin(theta), np.cos(theta)
        rotation_matrix = cls.rotation_matrix(sp, cp, st, ct)
        # Unit vectors
        zhat = rotation_matrix[2]
        v_dot_zhat = velocity @ zhat
        v_in_plane = velocity - zhat * v_dot_zhat
        angle_of_attack = -np.arctan(v_dot_zhat / np.linalg.norm(v_in_plane))
        return (
            angle_of_attack,
            sp,
            cp,
            st,
            ct,
            rotation_matrix,
            v_dot_zhat,
            v_in_plane,
        )

    def geometric_quantities(
        self,
        phi: float,
        theta: float,
        velocity: np.ndarray,
        angular_velocity: np.ndarray,
    ) -> _GeometricQuantitiesResult:
        # pylint: disable=too-many-locals
        """Compute various vectors and pseudo vectors from the euler angles.

        Compute intermediate quantities on the way to computing the time
        derivatives of the kinematic variables.
        """
        (
            angle_of_attack,
            _,
            _,
            st,
            ct,
            rotation_matrix,
            _,
            v_in_plane,
        ) = self.compute_angle_of_attack(phi, theta, velocity)
        # TODO: Does not guard against v_is_in_plane ~0
        # SEE: https://github.com/crzdg/flying-discs/pull/16#discussion_r2991607316
        zhat: np.ndarray = rotation_matrix[2]
        xhat: np.ndarray = v_in_plane / np.linalg.norm(v_in_plane)
        yhat = np.cross(zhat, xhat)
        w_prime = np.array(
            [
                angular_velocity[0] * ct,
                angular_velocity[1],
                angular_velocity[0] * st + angular_velocity[2],
            ]
        )
        # Angular velocity in lab coordinates
        w_lab = w_prime @ rotation_matrix
        # Angular velocity components along the unit vectors in the lab frame
        w = np.array([xhat, yhat, zhat]) @ w_lab
        return _GeometricQuantitiesResult(
            xhat=xhat,
            yhat=yhat,
            zhat=zhat,
            angle_of_attack=angle_of_attack,
            rotation_matrix=rotation_matrix,
            w_prime=w_prime,
            w_lab=w_lab,
            w=w,
        )

    def compute_forces(
        self,
        velocity: np.ndarray,
        geometric_quantities: _GeometricQuantitiesResult,
    ) -> _ForcesResult:
        """Computes the lift, drag, and gravitational forces on the disc."""
        # TODO: Does not handle || velocity || = 0.
        # SEE: https://github.com/crzdg/flying-discs/pull/16#discussion_r2991607198
        vhat = velocity / np.linalg.norm(velocity)
        force_amplitude = self.force_per_v2 * (velocity @ velocity)
        # Compute the lift and drag forces
        res = _ForcesResult(
            lift=self.model.lift(geometric_quantities.angle_of_attack)
            * force_amplitude
            * np.cross(vhat, geometric_quantities.yhat),
            drag=self.model.drag(geometric_quantities.angle_of_attack) * force_amplitude * (-vhat),
            grav=self.grav,
            mass=self.mass,
        )
        return res

    def compute_torques(
        self,
        velocity: np.ndarray,
        geometric_quantities: _GeometricQuantitiesResult,
    ) -> _TorquesResult:
        """Computes the torque around each principle axis."""
        torque_amplitude = self.torque_per_v2 * (velocity @ velocity)
        wx, wy, wz = geometric_quantities.w
        # Compute component torques. Note that "x" and "y" are computed
        # in the lab frame
        t_x_lab = self.model.x(wx, wz) * torque_amplitude * geometric_quantities.xhat
        t_y_lab = self.model.y(geometric_quantities.angle_of_attack, wy) * torque_amplitude * geometric_quantities.yhat
        z = self.model.z(wz) * torque_amplitude * self.z_hat
        x = geometric_quantities.rotation_matrix @ t_x_lab
        y = geometric_quantities.rotation_matrix @ t_y_lab
        return _TorquesResult(x_lab=t_x_lab, y_lab=t_y_lab, z=z, x=x, y=y)

    def compute_derivatives(self, _: float, coordinates: np.ndarray) -> np.ndarray:
        """Right hand side of the ordinary differential equations.

        This is supplied to :meth:`scipy.integrate.solve_ivp`. See `this page
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html#scipy.integrate.solve_ivp>`_
        for more information about its `fun` argument.

        .. todo::

           Implement the disc hitting the ground as a (Callable) scipy
           event object.

        Args:
          coordinates (np.ndarray): kinematic variables of the disc

        Returns:
          derivatives of all coordinates
        """
        # If the disk hit the ground, then stop calculations
        if coordinates[2] <= 0:
            return coordinates * 0

        velocity = coordinates[3:6]
        ang_velocity = coordinates[9:12]
        geometric_quantities = self.geometric_quantities(coordinates[6], coordinates[7], velocity, ang_velocity)
        forces = self.compute_forces(velocity, geometric_quantities)
        torques = self.compute_torques(velocity, geometric_quantities)
        return np.concatenate((velocity, forces.acc, ang_velocity, torques.total))



from __future__ import annotations

import math
from dataclasses import dataclass

from flying_discs.morrison.constants import Constants
from flying_discs.morrison.coordinates import MorrisonPosition2D, MorrisonTrajectory2D


@dataclass
class MorrisonBaseThrow:
    trajectory: MorrisonTrajectory2D
    constants: Constants
    z0: float
    v0: float
    angle_of_attack: float
    deltaT: float


class MorrisonBaseCalculator:
    def __init__(self, constants: Constants) -> None:
        self.constants = constants

    def calculate_trajectory_step(
        self, x: float, z: float, vx: float, vz: float, CD: float, CL: float, deltaT: float
    ) -> MorrisonPosition2D:
        ax = -self.constants.RHO * math.pow(vx, 2) * self.constants.A * CD * deltaT
        az = (
            self.constants.RHO * math.pow(vx, 2) * self.constants.A * CL / 2 / self.constants.MASS
            + self.constants.GRAVITY
        ) * deltaT
        next_vx = vx + ax
        next_vz = vz + az
        next_x = x + next_vx * deltaT
        next_z = z + next_vz * deltaT
        return MorrisonPosition2D(
            x=next_x,
            z=next_z,
            vx=next_vx,
            vz=next_vz,
            ax=ax,
            az=az,
        )

    def calculate_trajectory(
        self,
        z0: float,
        v0: float,
        angle_of_attack: float,
        deltaT: float,
    ) -> MorrisonBaseThrow:
        CL = self.constants.CL(alpha=angle_of_attack)
        CD = self.constants.CD(alpha=angle_of_attack)
        trajectory = [MorrisonPosition2D(z=z0, vx=v0)]
        while trajectory[-1].z > 0:
            current_step = trajectory[-1]
            next_position = self.calculate_trajectory_step(
                current_step.x, current_step.z, current_step.vx, current_step.vz, CD, CL, deltaT
            )
            trajectory.append(next_position)
        return MorrisonBaseThrow(
            MorrisonTrajectory2D(trajectory),
            self.constants,
            z0,
            v0,
            angle_of_attack,
            deltaT,
        )



from copy import deepcopy
from dataclasses import dataclass, field
from typing import List, Optional, Sequence, Tuple

import numpy as np

from flying_discs.morrison.constants import Constants
from flying_discs.morrison.coordinates import MorrisonPosition3D, MorrisonTrajectory3D
from flying_discs.morrison.linear import MorrisonLinearCalculator
from flying_discs.utils import quadratic_bezier, rotate_points_around_mid_point


@dataclass
class MorrisonBezierThrowExtra:
    bezier_points: List[Tuple[float, float]]
    linear_trajectory: MorrisonTrajectory3D
    intersect: List[Tuple[float, float]]


@dataclass
class MorrisonBezierThrow:
    # pylint: disable=too-many-instance-attributes
    trajectory: MorrisonTrajectory3D
    constants: Constants
    initial_position: MorrisonPosition3D
    v0: float
    angle_of_attack: float
    direction_angle: float
    intersect_angle: float
    factor: float
    deltaT: float
    # make an exception for the compare for the extras for easier testing
    # TODO: remove compare=False
    extras: Optional[MorrisonBezierThrowExtra] = field(default=None, compare=False)
    target_x: Optional[float] = None
    target_y: Optional[float] = None


class MorrisonBezierCalculator:
    def __init__(self, constants: Constants) -> None:
        self._linear_calculator = MorrisonLinearCalculator(constants)
        self.constants = constants

    @staticmethod
    def _determine_bezier_points(
        trajectory: Sequence[MorrisonPosition3D],
        rotated_points: List[Tuple[float, float]],
        tmax: int,
        factor: float,
    ) -> List[Tuple[float, float]]:
        x0 = trajectory[0].x
        y0 = trajectory[0].y
        idx = np.abs(int(tmax * factor))
        x1 = rotated_points[idx][0]
        y1 = rotated_points[idx][1]
        x2 = trajectory[-1].x
        y2 = trajectory[-1].y
        bezier_points = [(x0, y0), (x1, y1), (x2, y2)]
        return bezier_points

    def _update_trajectory_to_bezier(
        self, trajectory: MorrisonTrajectory3D, intersect_angle: float, factor: float
    ) -> Tuple[MorrisonTrajectory3D, MorrisonBezierThrowExtra]:
        updated_positions = deepcopy(trajectory.positions)
        intersect = rotate_points_around_mid_point(
            [(p.x, p.y) for p in updated_positions],
            intersect_angle,
        )

        tmax = len(updated_positions) - 1
        clipped_factor = min(1, max(-1, (factor + 1) / 2))
        bezier_points = self._determine_bezier_points(
            updated_positions,
            intersect,
            tmax,
            clipped_factor,
        )
        for t in range(tmax):
            x, y = quadratic_bezier(
                t / tmax,
                bezier_points[0],
                bezier_points[1],
                bezier_points[2],
            )
            updated_positions[t].x = x
            updated_positions[t].y = y

        return MorrisonTrajectory3D(updated_positions), MorrisonBezierThrowExtra(bezier_points, trajectory, intersect)

    def calculate_trajectory(
        self,
        initial_position: MorrisonPosition3D,
        v0: float,
        angle_of_attack: float,
        direction_angle: float,
        intersect_angle: float,
        factor: float,
        deltaT: float,
    ) -> MorrisonBezierThrow:
        linear_throw = self._linear_calculator.calculate_trajectory(
            initial_position, v0, angle_of_attack, direction_angle, deltaT
        )
        trajectory, extras = self._update_trajectory_to_bezier(linear_throw.trajectory, intersect_angle, factor)
        return MorrisonBezierThrow(
            trajectory,
            self.constants,
            initial_position,
            v0,
            angle_of_attack,
            direction_angle,
            intersect_angle,
            factor,
            deltaT,
            extras,
        )

    def calculate_trajectory_to_position(
        self,
        initial_position: MorrisonPosition3D,
        angle_of_attack: float,
        intersect_angle: float,
        factor: float,
        target_x: float,
        target_y: float,
        deltaT: float,
    ) -> MorrisonBezierThrow:
        linear_throw = self._linear_calculator.calculate_trajectory_to_position(
            initial_position, angle_of_attack, target_x, target_y, deltaT
        )
        trajectory, extras = self._update_trajectory_to_bezier(linear_throw.trajectory, intersect_angle, factor)
        return MorrisonBezierThrow(
            trajectory,
            self.constants,
            initial_position,
            linear_throw.v0,
            angle_of_attack,
            linear_throw.direction_angle,
            intersect_angle,
            factor,
            deltaT,
            extras,
            target_x,
            target_y,
        )



from typing import Any

import diffrax
import equinox as eqx
import jax.numpy as jnp
from diffrax import Solution

from flying_discs.frispy.jax_backend.eom import EOM
from flying_discs.frispy.jax_backend.model import Model


class Disc(eqx.Module):
    # pylint: disable=too-many-instance-attributes
    """Flying spinning disc object. Diffrax/JAX implementation."""

    area: float
    I_xx: float
    I_zz: float
    mass: float
    air_density: float
    g: float
    model: Model
    eom: EOM
    _term: diffrax.ODETerm
    _solver: diffrax.AbstractSolver = eqx.field(static=True)
    _stepsize_controller: diffrax.AbstractStepSizeController = eqx.field(static=True)

    def __init__(
        self,
        area: float = 0.058556,
        I_xx: float = 0.001219,
        I_zz: float = 0.002352,
        mass: float = 0.175,
        air_density: float = 1.225,
        g: float = 9.81,
    ) -> None:
        # pylint: disable=duplicate-code
        self.area = area
        self.I_xx = I_xx
        self.I_zz = I_zz
        self.mass = mass
        self.air_density = air_density
        self.g = g
        self.model = Model()
        self.eom = EOM(
            area=self.area,
            I_xx=self.I_xx,
            I_zz=self.I_zz,
            mass=self.mass,
            air_density=self.air_density,
            g=self.g,
        )
        self._term = diffrax.ODETerm(self.eom.compute_derivatives)  # type: ignore
        self._solver = diffrax.Dopri5()
        self._stepsize_controller = diffrax.PIDController(rtol=1e-3, atol=1e-6)

    # pylint: disable=duplicate-code
    def compute_trajectory(
        self,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 1.0,
        vx: float = 10.0,
        vy: float = 0.0,
        vz: float = 0.0,
        phi: float = 0.0,
        theta: float = 0.0,
        gamma: float = 0.0,
        dphi: float = 0.0,
        dtheta: float = 0.0,
        dgamma: float = 62.0,
        flight_time: float = 3.0,
        n_times: int = 100,
        **solver_kwargs: Any,
    ) -> tuple[dict[str, jnp.ndarray], Solution]:
        # pylint: disable=too-many-positional-arguments, too-many-arguments, too-many-locals
        """Call the Diffrax differential equation solver to compute the trajectory."""

        t_span = solver_kwargs.pop("t_span", (0.0, flight_time))
        t_eval = solver_kwargs.pop("t_eval", jnp.linspace(t_span[0], t_span[1], n_times))

        y0 = jnp.array(
            [
                x,
                y,
                z,
                vx,
                vy,
                vz,
                phi,
                theta,
                gamma,
                dphi,
                dtheta,
                dgamma,
            ]
        )

        result = self._solve(t_span, t_eval, y0)
        ts: jnp.ndarray = result.ts  # type: ignore
        ys: jnp.ndarray = result.ys  # type: ignore
        return (
            {
                "times": ts,
                "x": ys[:, 0],
                "y": ys[:, 1],
                "z": ys[:, 2],
                "vx": ys[:, 3],
                "vy": ys[:, 4],
                "vz": ys[:, 5],
                "phi": ys[:, 6],
                "theta": ys[:, 7],
                "gamma": ys[:, 8],
                "dphi": ys[:, 9],
                "dtheta": ys[:, 10],
                "dgamma": ys[:, 11],
            },
            result,
        )

    @eqx.filter_jit
    def _solve(self, t_span: jnp.ndarray, t_eval: jnp.ndarray, y0: jnp.ndarray) -> Solution:
        # TODO: Fix types for ODETerm, works for now
        # term = diffrax.ODETerm(self.eom.compute_derivatives)  # type: ignore

        saveat = diffrax.SaveAt(ts=t_eval)

        result: Solution = diffrax.diffeqsolve(
            self._term,
            self._solver,
            t0=t_span[0],
            t1=t_span[1],
            dt0=None,
            y0=y0,
            saveat=saveat,
            stepsize_controller=self._stepsize_controller,
            args=None,
        )
        return result



from dataclasses import dataclass
from typing import Any, TypedDict

import jax
import numpy as np

from flying_discs.frispy.constants import Constants
from flying_discs.frispy.coordinates import FrispyPosition, FrispyTrajectory
from flying_discs.frispy.jax_backend.disc import Disc as JAXDisc
from flying_discs.frispy.scipy_backend.disc import Disc


class ODEKwargs(TypedDict):
    x: float
    y: float
    z: float
    vx: float
    vy: float
    vz: float
    phi: float
    theta: float
    gamma: float
    dphi: float
    dtheta: float
    dgamma: float
    flight_time: float
    n_times: int


@dataclass
class FrispyThrow:
    # pylint: disable=too-many-instance-attributes
    trajectory: FrispyTrajectory
    constans: Constants
    initial_position: FrispyPosition
    vx: float
    vy: float
    vz: float
    phi: float
    theta: float
    gamma: float
    dphi: float
    dtheta: float
    dgamma: float
    deltaT: float


class FrispyCalculator:
    def __init__(self, constants: Constants, use_jax: bool = False, use_gpu: bool = False) -> None:
        self.constants = constants
        self._use_jax = use_jax

        if use_gpu and not use_jax:
            raise ValueError("GPU acceleration is only available with JAX. Set use_jax=True to use GPU.")
        self._use_gpu = use_gpu

        if not self._use_jax:
            self._disc = Disc(
                area=constants.AREA,
                I_xx=constants.I_xx,
                I_zz=constants.I_zz,
                mass=constants.MASS,
                air_density=constants.RH0,
                g=constants.GRAVITY,
            )

        else:

            base_disc = JAXDisc(
                area=constants.AREA,
                I_xx=constants.I_xx,
                I_zz=constants.I_zz,
                mass=constants.MASS,
                air_density=constants.RH0,
                g=constants.GRAVITY,
            )

            if self._use_gpu:
                try:
                    target_device = jax.devices("gpu")[0]
                except RuntimeError as exc:
                    raise RuntimeError("GPU requested, but JAX cannot find a valid CUDA device.") from exc
            else:
                target_device = jax.devices("cpu")[0]

            # Lock the object to the chosen hardware
            self._disc = jax.device_put(base_disc, target_device)

    def calculate_trajectory(
        # pylint: disable=too-many-arguments, too-many-locals, too-many-positional-arguments
        self,
        initial_position: FrispyPosition,
        vx0: float,
        vy0: float,
        vz0: float,
        phi: float,
        theta: float,
        gamma: float,
        dphi: float,
        dtheta: float,
        dgamma: float,
        deltaT: float,
    ) -> FrispyTrajectory:

        flight_time = 20.0
        n_times = int(1 / deltaT * flight_time)

        kwargs: ODEKwargs = {
            "x": initial_position.x,
            "y": initial_position.y,
            "z": initial_position.z,
            "vx": vx0,
            "vy": vy0,
            "vz": vz0,
            "phi": phi,
            "theta": theta,
            "gamma": gamma,
            "dphi": dphi,
            "dtheta": dtheta,
            "dgamma": dgamma,
            "flight_time": flight_time,
            "n_times": n_times,
        }

        solver_kwargs: dict[str, Any] = {}

        if self._use_jax and isinstance(self._disc, JAXDisc):
            jax_result, _ = self._disc.compute_trajectory(**kwargs)
            result = {k: np.asarray(v) for k, v in jax_result.items()}
        else:
            scipy_result, _ = self._disc.compute_trajectory(**kwargs, **solver_kwargs)
            result = scipy_result

        times = np.asarray(result["times"])

        positions = []
        for i in range(len(times)):
            positions.append(
                FrispyPosition(
                    x=result["x"][i],
                    y=result["y"][i],
                    z=result["z"][i],
                    vx=result["vx"][i],
                    vy=result["vy"][i],
                    vz=result["vz"][i],
                    phi=result["phi"][i],
                    theta=result["theta"][i],
                    gamma=result["gamma"][i],
                    dphi=result["dphi"][i],
                    dtheta=result["dtheta"][i],
                    dgamma=result["dgamma"][i],
                )
            )

        return FrispyTrajectory(positions)



import equinox as eqx
import jax.numpy as jnp


class Model(eqx.Module):
    # pylint: disable=too-many-instance-attributes
    """Kinematic model for a disc."""

    PL0: float = 0.33
    PLa: float = 1.9
    PD0: float = 0.18
    PDa: float = 0.69
    PTxwx: float = -0.013
    PTxwz: float = -0.0017
    PTy0: float = -0.082
    PTya: float = 0.43
    PTywy: float = -0.014
    PTzwz: float = -0.000034
    alpha_0: float = 4 * jnp.pi / 180

    def lift(self, alpha: jnp.ndarray) -> jnp.ndarray:
        return self.PL0 + self.PLa * alpha

    def drag(self, alpha: jnp.ndarray) -> jnp.ndarray:
        return self.PD0 + self.PDa * (alpha - self.alpha_0) ** 2

    def x(self, wx: float, wz: float) -> float:
        return self.PTxwx * wx + self.PTxwz * wz

    def y(self, alpha: jnp.ndarray, wy: float) -> jnp.ndarray:
        return self.PTy0 + self.PTywy * wy + self.PTya * alpha

    def z(self, wz: float) -> float:
        return self.PTzwz * wz



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

import numpy as np


@dataclass
class MorrisonPosition2D:
    def __init__(
        self,
        *,
        x: float = 0.0,
        z: float = 0.0,
        vx: float = 0.0,
        vz: float = 0.0,
        ax: float = 0.0,
        az: float = 0.0,
    ):
        self.x = x
        self.z = z
        self.vx = vx
        self.vz = vz
        self.ax = ax
        self.az = az


class MorrisonTrajectory2D:
    # pylint: disable=too-many-instance-attributes, invalid-name
    def __init__(self, positions: Sequence[MorrisonPosition2D]) -> None:
        self.positions = positions
        self.X = np.array([p.x for p in self.positions])
        self.Z = np.array([p.z for p in self.positions])
        self.VX = np.array([p.vx for p in self.positions])
        self.VZ = np.array([p.vz for p in self.positions])
        self.AX = np.array([p.ax for p in self.positions])
        self.AZ = np.array([p.az for p in self.positions])
        self.__n = 0  # iterator count

    def __iter__(self) -> "MorrisonTrajectory2D":
        self.__n = 0
        return self

    def __next__(self) -> MorrisonPosition2D:
        if self.__n <= len(self.positions) - 1:
            result = self.positions[self.__n]
            self.__n += 1
            return result
        raise StopIteration

    def __getitem__(self, i: int) -> MorrisonPosition2D:
        return self.positions[i]

    def __len__(self) -> int:
        return len(self.positions)

    def __eq__(self, o: object) -> bool:
        return self.positions == o


@dataclass
class MorrisonPosition3D:
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 0.0,
        vx: float = 0.0,
        vy: float = 0.0,
        vz: float = 0.0,
        ax: float = 0.0,
        ay: float = 0.0,
        az: float = 0.0,
    ):
        self.x = x
        self.y = y
        self.z = z
        self.vx = vx
        self.vy = vy
        self.vz = vz
        self.ax = ax
        self.ay = ay
        self.az = az


class MorrisonTrajectory3D:
    # pylint: disable=too-many-instance-attributes, invalid-name

    def __init__(self, positions: Sequence[MorrisonPosition3D]) -> None:
        self.positions = positions
        self.X = np.array([p.x for p in self.positions])
        self.Y = np.array([p.y for p in self.positions])
        self.Z = np.array([p.z for p in self.positions])
        self.VX = np.array([p.vx for p in self.positions])
        self.VY = np.array([p.vy for p in self.positions])
        self.VZ = np.array([p.vz for p in self.positions])
        self.AX = np.array([p.ax for p in self.positions])
        self.AY = np.array([p.ay for p in self.positions])
        self.AZ = np.array([p.az for p in self.positions])
        self.__n = 0  # iterator count

    def __iter__(self) -> "MorrisonTrajectory3D":
        self.__n = 0
        return self

    def __next__(self) -> MorrisonPosition3D:
        if self.__n <= len(self.positions) - 1:
            result = self.positions[self.__n]
            self.__n += 1
            return result
        raise StopIteration

    def __getitem__(self, i: int) -> MorrisonPosition3D:
        return self.positions[i]

    def __len__(self) -> int:
        return len(self.positions)

    def __eq__(self, o: object) -> bool:
        return self.positions == o
